# Combine version map fragments into version scripts for our shared objects.
# Copyright (C) 1998,99,2000 Free Software Foundation, Inc.
# Written by Ulrich Drepper <drepper@cygnus.com>, 1998.

# This script expects the following variables to be defined:
# defsfile		name of Versions.def file
# srcroot		name of source root directory with trailing slash
# buildroot		name of build directory with trailing slash
# move_if_change	move-if-change command

# Read definitions for the versions.
BEGIN {
  nlibs=0;
  while (getline < defsfile) {
    if (/^[a-zA-Z0-9_.]+ \{/) {
      libs[$1] = 1;
      curlib = $1;
      while (getline < defsfile && ! /^}/) {
      if ($2 == "=") {
	  renamed[curlib "::" $1] = $3;
      }
	else
	  versions[$1] = 1;
      }
    }
  }
  close(defsfile);

  tmpfile = buildroot "Versions.tmp";
  sort = "sort -n > " tmpfile;
}

# Remove comment lines.
/^ *#/ {
  next;
}

# This matches the beginning of the version information for a new library.
/^[a-zA-Z0-9_.]+/ {
  actlib = $1;
  if (!libs[$1]) {
    printf("no versions defined for %s\n", $1) > "/dev/stderr";
    exit 1;
  }
  next;
}

# This matches the beginning of a new version for the current library.
/^  [A-Za-z_]/ {
  if (renamed[actlib "::" $1])
    actver = renamed[actlib "::" $1];
  else if (!versions[$1]) {
    printf("version %s not defined\n", $1) > "/dev/stderr";
    exit 1;
  }
  else
    actver = $1;
  next;
}

# This matches lines with names to be added to the current version in the
# current library.  This is the only place where we print something to
# the intermediate file.
/^   / {
  printf("%s %s %s\n", actlib, actver, $0) | sort;
}


function closeversion(name, oldname) {
  if (firstinfile) {
    printf("  local:\n    *;\n") > outfile;
    firstinfile = 0;
  }
  printf("}%s;\n", oldname) > outfile;
}

function close_and_move(name, real_name) {
  close(name);
  system(move_if_change " " name " " real_name " >&2");
}

# Now print the accumulated information.
END {
  system("mkdir -p " buildroot "bits");
  real_symver = buildroot "bits/symver.h.new"
  symver = real_symver "T"
  system("cp " srcroot "misc/bits/symver.h.in " symver);
  system("chmod +w " symver);
  real_symverinfo = buildroot "bits/symverinfo.h.new"
  symverinfo = real_symverinfo "T"
  system("cp " srcroot "misc/bits/symverinfo.h.in " symverinfo);
  system("chmod +w " symverinfo);
  # We exclude some functions like system call wrappers in libpthread.
  excludes="accept; close; connect; fcntl; fsync; lseek; lseek64; \
	    msync; nanosleep; open; open64; pause; pread; pread64; \
	    pwrite; pwrite64; read; recv; recvfrom; recvmsg; send; \
	    sendmsg; sendto; system; tcdrain; wait; waitpid; write;";
  split(excludes, excluded);
  close(sort);
  oldlib = "";
  oldver = "";
  printf("version-maps =");
  while(getline < tmpfile) {
    if ($1 != oldlib) {
      if (oldlib != "") {
	closeversion(oldver, veryoldver);
	oldver = "";
	close_and_move(outfile, real_outfile);
      }
      oldlib = $1;
      real_outfile = buildroot oldlib ".map";
      outfile = real_outfile "T";
      firstinfile = 1;
      veryoldver = "";
      printf(" %s.map", oldlib);
    }
    if ($2 != oldver) {
      if (oldver != "") {
	closeversion(oldver, veryoldver);
	veryoldver = oldver;
      }
      printf("%s {\n  global:\n", $2) > outfile;
      oldver = $2;
    }
    printf("   ") > outfile;
    for (n = 3; n <= NF; ++n) {
      printf(" %s", $n) > outfile;
      if ($n ~ /^_/ && $n !~ /^_sys_errlist/ && $n !~ /^_sys_siglist/ && $n !~ /^_sys_nerr/)
	continue;
      # No pthread functions.
      if ($n ~ /^pthread_/ && $n !~ /^pthread_attr_init/)
	continue;
      if ($n ~ /[*\.\[\]\?]/)
	continue;
      skip=0;
      for (i in excluded) {
	if ($n ~ "^" excluded[i]) {
	  skip=1;
	  break;
	}
      }
      if (skip == 1)
	continue;
      if ($1 == "libc") {
	symbol=$n;
	sub (/;/, "", symbol);
	symbols[$2, symbol] = symbol;
	nsymbols[symbol]++;
      }
    }
    printf("\n") > outfile;
  }
  close(tmpfile);
  printf("\n");
  closeversion(oldver, veryoldver);
  close_and_move(outfile, real_outfile);
  printf("#ifdef __USE_VERSIONED_INTERFACE\n\n") >> symver;
  for (v in versions) {
    printf ("%s\n", v) | sort;
  }
  close(sort);
  totalversion=0;
  while(getline < tmpfile) {
    sortedversions[totalversion]=$1;
    totalversion++;
  }
  close(tmpfile)
  for (s in nsymbols) {
    printf ("%s\n", s) | sort;
    nsymbols[s]=0;
    for (v in versions) {
      if (symbols[v, s]) {
	nsymbols[s]++;
      }
    }
  }
  close(sort);
  while(getline < tmpfile) {
    s=$1;
    if (nsymbols[s] > 1) {
      nversions=0;
      for (v in versions) {
	if (symbols[v, s]) {
	  nversions++;
	}
      }
      for (j = 0; j < totalversion; j++) {
	if (symbols[sortedversions[j], s]) {
	  n=j;
	  for (i = n + 1; i < totalversion; i++) {
	    if (symbols[sortedversions[i], s]) {
	      n=i;
	      break;
	    }
	  }
	  m="__USE_" sortedversions[j] "_" s;
	  gsub (/\./, "_", m);
	  for (i = j; i < n; i++) {
	    glibcver="__USE_" sortedversions[i];
	    gsub (/\./, "_", glibcver);
	    printf("#ifdef %s\n", glibcver) >> symverinfo
	    printf("#define %s\n", m) >> symverinfo
	    printf("#endif\n") >> symverinfo
	  }
        }
      }
      i=0;
      for (j = 0; j < totalversion; j++) {
	v=sortedversions[j];
	if (symbols[v, s]) {
	  i++;
	  m="__USE_" v "_" s;
	  gsub (/\./, "_", m);
	  if (i == 1) {
	    printf("#ifdef %s\n", m) >> symver;
	    printf("_symbol_version (%s, %s, %s);\n", s, s, v) >> symver;
	  }
	  else if (i != nversions) {
	    printf("#elif defined %s\n", m) >> symver;
	    printf("_symbol_version (%s, %s, %s);\n", s, s, v) >> symver;
	  }
	  else {
	    printf("#else\n") >> symver;
	    printf("_symbol_version (%s, %s, %s);\n", s, s, v) >> symver;
	    printf("#endif\n") >> symver;
	  }
	}
      }
    }
    else {
      for (v in versions) {
	if (symbols[v, s]) {
	  printf("_full_symbol_version (%s, %s, %s);\n", s, s, v) >> symver;
	  break;
	}
      }
    }
  }
  close(tmpfile)
  printf("\n#endif /* !__USE_VERSIONED_INTERFACE */\n") >> symver;
  printf("\n#endif /* bits/symver.h */\n") >> symver;
  close_and_move(symver, real_symver);
  printf("\n#endif /* bits/symverinfo.h */\n") >> symverinfo;
  close_and_move(symverinfo, real_symverinfo);
  system("rm -f " tmpfile);
}
